package dev.inmo.postssystem.server

import dev.inmo.postssystem.features.auth.server.AuthenticationRoutingConfigurator
import dev.inmo.postssystem.features.auth.server.SessionAuthenticationConfigurator
import dev.inmo.postssystem.features.auth.server.tokens.*
import dev.inmo.postssystem.features.common.common.getAllDistinct
import dev.inmo.postssystem.features.common.common.singleWithBinds
import dev.inmo.postssystem.features.common.server.sessions.ApplicationAuthenticationConfigurator
import dev.inmo.postssystem.features.files.common.*
import dev.inmo.postssystem.features.files.common.storage.*
import dev.inmo.postssystem.features.files.common.storage.WriteFilesStorage
import dev.inmo.postssystem.features.files.server.FilesRoutingConfigurator
import dev.inmo.postssystem.features.roles.common.*
import dev.inmo.postssystem.features.roles.common.keyvalue.KeyValuesUsersRolesOriginalRepo
import dev.inmo.postssystem.features.roles.manager.common.RolesManagerRole
import dev.inmo.postssystem.features.roles.manager.common.RolesManagerRoleStorage
import dev.inmo.postssystem.features.roles.manager.server.RolesManagerRolesChecker
import dev.inmo.postssystem.features.roles.manager.server.RolesManagerUsersRolesStorageServerRoutesConfigurator
import dev.inmo.postssystem.features.roles.server.*
import dev.inmo.postssystem.features.status.server.StatusRoutingConfigurator
import dev.inmo.postssystem.features.users.common.ExposedUsersStorage
import dev.inmo.postssystem.features.users.server.UsersStorageServerRoutesConfigurator
import dev.inmo.micro_utils.coroutines.LinkedSupervisorScope
import dev.inmo.micro_utils.ktor.server.configurators.*
import dev.inmo.micro_utils.ktor.server.createKtorServer
import dev.inmo.micro_utils.repos.exposed.keyvalue.ExposedKeyValueRepo
import dev.inmo.micro_utils.repos.exposed.onetomany.ExposedOneToManyKeyValueRepo
import io.ktor.application.featureOrNull
import io.ktor.application.log
import io.ktor.routing.Route
import io.ktor.routing.Routing
import io.ktor.server.engine.ApplicationEngine
import io.ktor.server.netty.Netty
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.serialization.StringFormat
import kotlinx.serialization.json.Json
import org.jetbrains.exposed.sql.Database
import org.koin.core.module.Module
import org.koin.core.parameter.ParametersHolder
import org.koin.core.qualifier.StringQualifier
import org.koin.dsl.*
import java.io.File

private fun Route.print() {
    println(this)
    children.forEach {
        it.print()
    }
}


fun getDIModule(
    vararg args: String,
    baseScope: CoroutineScope = CoroutineScope(Dispatchers.IO)
): Module {
    val json = Json {
        ignoreUnknownKeys = true
    }
    val config = json.decodeFromString(Config.serializer(), File(args.first()).readText())

    val originalFilesMetasKeyValueRepoQualifier = StringQualifier("OriginalFilesMetaKV")
    val filesMetasKeyValueRepoQualifier = StringQualifier("FilesMetaKV")
    val filesFolderQualifier = StringQualifier("filesFolder")
    val reviewVersionsQualifier = StringQualifier("reviewVersions")
    val releaseVersionsQualifier = StringQualifier("releaseVersions")
    val usersRolesKeyValueFactoryQualifier = StringQualifier("usersRolesKeyValueFactory")

    return module {
        single { json }
        single<StringFormat> { get<Json>() }
        singleWithBinds { config }
        singleWithBinds { get<Config>().databaseConfig }
        singleWithBinds { get<Config>().authConfig }
        singleWithBinds(filesFolderQualifier) { get<Config>().filesFolderFile }

        singleWithBinds { get<DatabaseConfig>().database }
        singleWithBinds(originalFilesMetasKeyValueRepoQualifier) {
            ExposedKeyValueRepo(get(), { text("fileid") }, { text("metaInfo") }, "FileIdsToMetas")
        }
        singleWithBinds(filesMetasKeyValueRepoQualifier) {
            MetasKeyValueRepo(
                get(),
                get(originalFilesMetasKeyValueRepoQualifier)
            )
        }
        single<FilesStorage> { DiskFilesStorage(get(filesFolderQualifier), get(filesMetasKeyValueRepoQualifier)) }
        single<WriteFilesStorage> { WriteDistFilesStorage(get(filesFolderQualifier), get(filesMetasKeyValueRepoQualifier)) }
        single<FullFilesStorage> { DefaultFullFilesStorage(get(), get()) }
        singleWithBinds { ExposedUsersStorage(get()) }
        singleWithBinds { exposedUsersAuthenticator(get(), get()) }

        factory<KeyValuesUsersRolesOriginalRepo>(usersRolesKeyValueFactoryQualifier) { (tableName: String) ->
            ExposedOneToManyKeyValueRepo(get(), { long("userId") }, { text("role") }, tableName)
        }
        single {
            RolesManagerRoleStorage(get(usersRolesKeyValueFactoryQualifier) { ParametersHolder(mutableListOf("rolesManager")) })
        }
        single<UsersRolesStorage<RolesManagerRole>>(StringQualifier("RolesManagerRoleStorage")) { get<RolesManagerRoleStorage>() }
        singleWithBinds {
            UsersRolesStorageHolder(
                RolesManagerRole::class,
                get<RolesManagerRoleStorage>()
            )
        }
        singleWithBinds<UsersRolesStorage<UserRole>> { UsersRolesAggregator(getAll()) }

        // Roles checkers
        single<RolesChecker<UserRole>>(StringQualifier(RolesManagerRolesChecker.key)) { RolesManagerRolesChecker }

        factory<CoroutineScope> { baseScope.LinkedSupervisorScope() }

        // Routing configurators
        singleWithBinds { FilesRoutingConfigurator(get(), null) }
        singleWithBinds { StatusRoutingConfigurator }
        singleWithBinds { UsersStorageServerRoutesConfigurator(get()) }
        singleWithBinds { UsersRolesStorageReadServerRoutesConfigurator<UserRole>(get(), UserRoleSerializer) }
        singleWithBinds { RolesManagerUsersRolesStorageServerRoutesConfigurator(get()) }

        singleWithBinds { ClientStaticRoutingConfiguration(get<Config>().clientStatic) }
        singleWithBinds {
            UsersRolesAuthenticationConfigurator<UserRole>(
                get(),
                get(),
                getAll()
            )
        }

        // Session and auth configurators
        singleWithBinds { SessionAuthenticationConfigurator(get<AuthConfig>().sessionAge) }
        singleWithBinds {
            DefaultAuthTokensService(
                get<Database>(),
                get(),
                get(),
                get<AuthConfig>().sessionAge,
                get()
            )
        }
        singleWithBinds { AuthenticationRoutingConfigurator(get(), get()) }

        if (config.debugMode) {
            single<ApplicationRoutingConfigurator.Element>(StringQualifier("Tracer")) { ApplicationRoutingConfigurator.Element {(this as Routing).trace { application.log.trace(it.buildText()) } } }
        }

        singleWithBinds { ApplicationRoutingConfigurator(getAllDistinct()) }
        singleWithBinds { ApplicationCachingHeadersConfigurator(getAllDistinct()) }
        singleWithBinds { ApplicationSessionsConfigurator(getAllDistinct()) }
        singleWithBinds { StatusPagesConfigurator(getAllDistinct()) }
        singleWithBinds { ApplicationAuthenticationConfigurator(getAllDistinct()) }
        singleWithBinds { WebSocketsConfigurator }

        factory<ApplicationEngine> {
            val config = get<Config>()
            createKtorServer(
                Netty,
                config.host,
                config.port
            ) {
                getAllDistinct<KtorApplicationConfigurator>().also(::println).forEach {
                    it.apply { configure() }
                }
                if (config.debugMode) {
                    featureOrNull(Routing) ?.print()
                }
            }
        }
    }
}
